use crate::flow::comparator::*;

#[derive(Clone)]
pub struct Bounds<A> {
    l: Option<A>,
    r: Option<A>,
}

pub enum BoundsError<A> {
    BoundsViolated { l: A, r: A },
    OverflowL(A),
    OverflowR(A),
}

impl<A: Clone> Bounds<A> {
    pub fn unbound() -> Self {
        Bounds { l: None, r: None }
    }

    fn ordered(l: &A, r: &A, comparator: &impl Comparator<A>) -> Result<(), BoundsError<A>> {
        if let Comparison::R = comparator.compare(l, r) {
            Err(BoundsError::BoundsViolated {
                l: l.clone(),
                r: r.clone(),
            })
        } else {
            Ok(())
        }
    }

    fn orderedl(
        l: &Option<A>,
        r: &A,
        comparator: &impl Comparator<A>,
    ) -> Result<(), BoundsError<A>> {
        if let Some(l) = &l {
            Self::ordered(l, r, comparator)
        } else {
            Err(BoundsError::OverflowL(r.clone()))
        }
    }

    fn orderedr(
        l: &A,
        r: &Option<A>,
        comparator: &impl Comparator<A>,
    ) -> Result<(), BoundsError<A>> {
        if let Some(r) = &r {
            Self::ordered(l, r, comparator)
        } else {
            Err(BoundsError::OverflowR(l.clone()))
        }
    }

    fn new(
        l: Option<A>,
        r: Option<A>,
        comparator: &impl Comparator<A>,
    ) -> Result<Self, BoundsError<A>> {
        if let (Some(kl), Some(kr)) = (&l, &r) {
            Self::ordered(kl, kr, comparator)?
        }
        Ok(Bounds { l, r })
    }

    pub fn split(
        self,
        key: &A,
        comparator: &impl Comparator<A>,
    ) -> Result<(Self, Self), BoundsError<A>> {
        Ok((
            Self::new(self.l, Some(key.clone()), comparator)?,
            Self::new(Some(key.clone()), self.r, comparator)?,
        ))
    }

    pub fn join(
        l: Self,
        r: Self,
        key: &A,
        comparator: &impl Comparator<A>,
    ) -> Result<Self, BoundsError<A>> {
        Self::orderedl(&l.r, key, comparator)?;
        Self::orderedr(key, &r.l, comparator)?;
        Self::new(l.l, r.r, comparator)
    }

    fn equal_bound(l: &Option<A>, r: &Option<A>, comparator: &impl Comparator<A>) -> bool {
        match (l, r) {
            (None, None) => true,
            (Some(kl), Some(kr)) => comparator.equal(kl, kr),
            _ => false,
        }
    }

    pub fn equal(bl: &Self, br: &Self, comparator: &impl Comparator<A>) -> bool {
        Self::equal_bound(&bl.l, &br.l, comparator) && Self::equal_bound(&bl.r, &br.r, comparator)
    }
}
