//! Add extra behaviour on top of an existing instance.
//!
//! Initially made as [`EmbedFail`] implementation detail.
//!
//! For combining existing [instances], see [`composition`].
//!
//! [`composition`]: super::composition

use std::marker::PhantomData;

use crate::func::class_prelude::*;

pub struct OverloadInstance<T, O>(T, O);

pub trait DeriveWeakFunctor: Send {}
impl<O: DeriveFunctor> DeriveWeakFunctor for O {}
pub trait DeriveFunctor: Send {}
impl<O: DeriveApplicative> DeriveFunctor for O {}
pub trait DeriveApplicative: Send {}
impl<O: DeriveMonad> DeriveApplicative for O {}
pub trait DeriveMonad: Send {}

impl<'a, T: WeakFunctor<'a>, O: 'a + DeriveWeakFunctor> WeakFunctor<'a> for OverloadInstance<T, O> {
    type F<A: 'a + Send> = T::F<A>;
}

impl<'a, T: Functor<'a>, O: 'a + DeriveFunctor> Functor<'a> for OverloadInstance<T, O> {
    fn fmap<A: 'a + Send, B: 'a + Send>(
        fa: Self::F<A>,
        f: impl 'a + Send + FnOnce(A) -> B,
    ) -> Self::F<B> {
        T::fmap(fa, f)
    }

    fn replace<A: 'a + Send, B: 'a + Send>(fa: Self::F<A>, b: B) -> Self::F<B> {
        T::replace(fa, b)
    }

    fn void<A: 'a + Send>(fa: Self::F<A>) -> Self::F<()> {
        T::void(fa)
    }
}

impl<'a, T: Pure<'a>, O: 'a + DeriveApplicative> Pure<'a> for OverloadInstance<T, O> {
    fn pure<A: 'a + Send>(a: A) -> Self::F<A> {
        T::pure(a)
    }
}

impl<'a, T: ApplicativeSeq<'a>, O: 'a + DeriveApplicative> ApplicativeSeq<'a>
    for OverloadInstance<T, O>
{
    fn seq<A: 'a + Send, B: 'a + Send>(
        ff: Self::F<impl 'a + Send + FnOnce(A) -> B>,
        fa: Self::F<A>,
    ) -> Self::F<B> {
        T::seq(ff, fa)
    }
}

impl<'a, T: ApplicativeLA2<'a>, O: 'a + DeriveApplicative> ApplicativeLA2<'a>
    for OverloadInstance<T, O>
{
    fn la2<A: 'a + Send, B: 'a + Send, C: 'a + Send>(
        fa: Self::F<A>,
        fb: Self::F<B>,
        f: impl 'a + Send + FnOnce(A, B) -> C,
    ) -> Self::F<C> {
        T::la2(fa, fb, f)
    }
}

impl<'a, T: ApplicativeTuple<'a>, O: 'a + DeriveApplicative> ApplicativeTuple<'a>
    for OverloadInstance<T, O>
{
    fn tuple<A: 'a + Send, B: 'a + Send>(fab: (Self::F<A>, Self::F<B>)) -> Self::F<(A, B)> {
        T::tuple(fab)
    }
}

impl<'a, T: ApplicativeSelect<'a>, O: 'a + DeriveApplicative> ApplicativeSelect<'a>
    for OverloadInstance<T, O>
{
    fn select<A: 'a + Send, B: 'a + Send>(
        fa: Self::F<A>,
        fb: Self::F<B>,
    ) -> SelectedWrapped<'a, A, B, Self> {
        T::select_map(fa, fb, |selected| match selected {
            Selected::A(a, fb) => Selected::A(a, fb),
            Selected::B(fa, b) => Selected::B(fa, b),
        })
    }
}

impl<'a, T: Applicative<'a>, O: 'a + DeriveApplicative> Applicative<'a> for OverloadInstance<T, O> {
    fn discard_first<A: 'a + Send, B: 'a + Send>(fa: Self::F<A>, fb: Self::F<B>) -> Self::F<B> {
        T::discard_first(fa, fb)
    }

    fn discard_second<A: 'a + Send, B: 'a + Send>(fa: Self::F<A>, fb: Self::F<B>) -> Self::F<A> {
        T::discard_second(fa, fb)
    }
}

struct OverloadIterative<F, O>(F, PhantomData<O>);

impl<F, O> OverloadIterative<F, O> {
    fn new(f: F) -> Self {
        Self(f, PhantomData)
    }
}

impl<'a, T: Monad<'a>, O: 'a + DeriveMonad, F: Iterative<'a, T = OverloadInstance<T, O>>>
    Iterative<'a> for OverloadIterative<F, O>
{
    type B = F::B;

    type T = T;

    fn next(self) -> IterativeWrapped<'a, Self> {
        T::fmap(self.0.next(), |state| {
            ControlFlow::Continue(Self::new(state?))
        })
    }
}

impl<'a, T: Monad<'a>, O: 'a + DeriveMonad> Monad<'a> for OverloadInstance<T, O> {
    fn bind<A: 'a + Send, B: 'a + Send>(
        fa: Self::F<A>,
        f: impl 'a + Send + FnOnce(A) -> Self::F<B>,
    ) -> Self::F<B> {
        T::bind(fa, f)
    }

    fn iterate<B: 'a + Send>(f: impl Iterative<'a, T = Self, B = B>) -> Self::F<B> {
        T::iterate(OverloadIterative::new(f))
    }

    fn join<A: 'a + Send>(ffa: Self::F<Self::F<A>>) -> Self::F<A> {
        T::join(ffa)
    }
}

pub struct DeriveFail<Ex>(Ex);

impl<Ex: Send> DeriveMonad for DeriveFail<Ex> {}

impl<'a, E: 'a + Send, Ex: 'a + Send, T: MonadFail<'a, Result<E, Ex>>> Fail<'a, E>
    for EmbedFail<T, Ex>
{
    fn fail<A: 'a + Send>(e: E) -> Self::F<A> {
        T::fail(Ok(e))
    }
}

/// Instance of [`MonadFailAny`] for [`EmbedFail`].
pub struct DeriveFailAny<Ex, Fallible>(Ex, Fallible);

impl<'a, Ex: 'a + Send, Fallible: MonadFailAny<'a>> MonadFailAny<'a>
    for DeriveFailAny<Ex, Fallible>
{
    type W<E: 'a + Send> = EmbedFail<Fallible::W<Result<E, Ex>>, Ex>;

    type T = Fallible::W<Ex>;

    fn unstuff<A: 'a + Send, E: 'a + Send>(
        wa: WrapE<'a, A, E, Self>,
    ) -> Wrap<'a, Result<A, E>, Self::T> {
        Fallible::bind_err(<Self::W<E> as Functor>::fmap(wa, Ok), |err| match err {
            Ok(e) => Fallible::pure(Err(e)),
            Err(ex) => Fallible::fail(ex),
        })
    }

    fn stuff<A: 'a + Send, E: 'a + Send>(
        fa: Wrap<'a, Result<A, E>, Self::T>,
    ) -> WrapE<'a, A, E, Self> {
        Fallible::bind(fa, |result| match result {
            Ok(Ok(a)) => Fallible::pure(a),
            Ok(Err(e)) => Fallible::fail(Ok(e)),
            Err(ex) => Fallible::fail(Err(ex)),
        })
    }

    fn map_err<A: 'a + Send, E0: 'a + Send, E1: 'a + Send>(
        wa: WrapE<'a, A, E0, Self>,
        f: impl 'a + Send + FnOnce(E0) -> E1,
    ) -> WrapE<'a, A, E1, Self> {
        Fallible::map_err(wa, |err| err.map(f))
    }

    fn bind_err<A: 'a + Send, E0: 'a + Send, E1: 'a + Send>(
        wa: WrapE<'a, A, E0, Self>,
        f: impl 'a + Send + FnOnce(E0) -> WrapE<'a, A, E1, Self>,
    ) -> WrapE<'a, A, E1, Self> {
        Fallible::bind_err(wa, |err| match err {
            Ok(e0) => f(e0),
            Err(ex) => Fallible::fail(Err(ex)),
        })
    }

    fn bind<A: 'a + Send, B: 'a + Send, E0: 'a + Send, E1: 'a + Send>(
        wa: WrapE<'a, A, E0, Self>,
        f: impl 'a + Send + FnOnce(Result<A, E0>) -> WrapE<'a, B, E1, Self>,
    ) -> WrapE<'a, B, E1, Self> {
        Fallible::bind(wa, |result| match result {
            Ok(a) => f(Ok(a)),
            Err(Ok(e0)) => f(Err(e0)),
            Err(Err(ex)) => Fallible::fail(Err(ex)),
        })
    }
}

impl<'a, T: SharedFunctor<'a>, O: 'a + DeriveWeakFunctor> SharedFunctor<'a>
    for OverloadInstance<T, O>
{
    type Shared<A: 'a + Send + Sync + Clone> = T::Shared<A>;

    fn share<A: 'a + Send + Sync + Clone>(fa: Self::F<A>) -> Self::Shared<A> {
        T::share(fa)
    }

    fn unshare<A: 'a + Send + Sync + Clone>(sa: Self::Shared<A>) -> Self::F<A> {
        T::unshare(sa)
    }
}

/// [`MonadFail`] based on embedding the error into some other [`MonadFailAny`].
pub type EmbedFail<T, Ex> = OverloadInstance<T, DeriveFail<Ex>>;
