//! Fallible version of [`future`].
//!
//! Preferred over building it out of [`composition`].
//!
//! To embed extra errors, see [`overload::EmbedFail`].
//!
//! [`future`]: super::future
//! [`composition`]: super::composition
//! [`overload::EmbedFail`]: super::overload::EmbedFail

use std::{future::Future, pin::Pin};

use futures::{
    future::{try_select, Either, Shared},
    try_join, FutureExt,
};

use crate::func::class_prelude::*;

pub struct TryFutureInstance<E>(E);

impl<E: Send> WeakFunctorAny for TryFutureInstance<E> {
    type FAny<'a, A: 'a + Send> = Pin<Box<dyn 'a + Send + Future<Output = Result<A, E>>>> where Self: 'a;
}

impl<'a, E: 'a + Send> Functor<'a> for TryFutureInstance<E> {
    fn fmap<A: 'a + Send, B: 'a + Send>(
        fa: Self::F<A>,
        f: impl 'a + Send + FnOnce(A) -> B,
    ) -> Self::F<B> {
        Box::pin(async { Ok(f(fa.await?)) })
    }

    fn replace<A: 'a + Send, B: 'a + Send>(fa: Self::F<A>, b: B) -> Self::F<B> {
        Box::pin(async {
            fa.await?;
            Ok(b)
        })
    }
}

impl<'a, E: 'a + Send> Pure<'a> for TryFutureInstance<E> {
    fn pure<A: 'a + Send>(a: A) -> Self::F<A> {
        Box::pin(async { Ok(a) })
    }
}

impl<'a, E: 'a + Send> ApplicativeSeq<'a> for TryFutureInstance<E> {
    fn seq<A: 'a + Send, B: 'a + Send>(
        ff: Self::F<impl 'a + Send + FnOnce(A) -> B>,
        fa: Self::F<A>,
    ) -> Self::F<B> {
        Box::pin(async {
            let (f, a) = try_join!(ff, fa)?;
            Ok(f(a))
        })
    }
}

impl<'a, E: 'a + Send> ApplicativeLA2<'a> for TryFutureInstance<E> {
    fn la2<A: 'a + Send, B: 'a + Send, C: 'a + Send>(
        fa: Self::F<A>,
        fb: Self::F<B>,
        f: impl 'a + Send + FnOnce(A, B) -> C,
    ) -> Self::F<C> {
        Box::pin(async {
            let (a, b) = try_join!(fa, fb)?;
            Ok(f(a, b))
        })
    }
}

impl<'a, E: 'a + Send> ApplicativeTuple<'a> for TryFutureInstance<E> {
    fn tuple<A: 'a + Send, B: 'a + Send>((fa, fb): (Self::F<A>, Self::F<B>)) -> Self::F<(A, B)> {
        Box::pin(async { try_join!(fa, fb) })
    }
}

impl<'a, E: 'a + Send> ApplicativeSelect<'a> for TryFutureInstance<E> {
    fn select<A: 'a + Send, B: 'a + Send>(
        fa: Self::F<A>,
        fb: Self::F<B>,
    ) -> SelectedWrapped<'a, A, B, Self> {
        Box::pin(async {
            match try_select(fa, fb).await {
                Ok(Either::Left((a, fb))) => Ok(Selected::A(a, fb)),
                Ok(Either::Right((b, fa))) => Ok(Selected::B(fa, b)),
                Err(Either::Left((e, _))) => Err(e),
                Err(Either::Right((e, _))) => Err(e),
            }
        })
    }
}

impl<'a, E: 'a + Send> Applicative<'a> for TryFutureInstance<E> {
    fn discard_first<A: 'a + Send, B: 'a + Send>(fa: Self::F<A>, fb: Self::F<B>) -> Self::F<B> {
        Box::pin(async { Ok(try_join!(fa, fb)?.1) })
    }

    fn discard_second<A: 'a + Send, B: 'a + Send>(fa: Self::F<A>, fb: Self::F<B>) -> Self::F<A> {
        Box::pin(async { Ok(try_join!(fa, fb)?.0) })
    }
}

impl<'a, E: 'a + Send> Monad<'a> for TryFutureInstance<E> {
    fn bind<A: 'a + Send, B: 'a + Send>(
        fa: Self::F<A>,
        f: impl 'a + Send + FnOnce(A) -> Self::F<B>,
    ) -> Self::F<B> {
        Box::pin(async { f(fa.await?).await })
    }

    fn iterate<B: 'a + Send>(mut f: impl Iterative<'a, T = Self, B = B>) -> Self::F<B> {
        Box::pin(async move {
            loop {
                match f.next().await? {
                    ControlFlow::Continue(next_f) => f = next_f,
                    ControlFlow::Break(b) => return Ok(b),
                }
            }
        })
    }

    fn join<A: 'a + Send>(ffa: Self::F<Self::F<A>>) -> Self::F<A> {
        Box::pin(async { ffa.await?.await })
    }
}

impl<'a, E: 'a + Send + Sync + Clone> SharedFunctor<'a> for TryFutureInstance<E> {
    type Shared<A: 'a + Send + Sync + Clone> =
        Shared<Pin<Box<dyn 'a + Send + Future<Output = Result<A, E>>>>>;

    fn share<A: 'a + Send + Sync + Clone>(fa: Self::F<A>) -> Self::Shared<A> {
        fa.shared()
    }

    fn unshare<A: 'a + Send + Sync + Clone>(sa: Self::Shared<A>) -> Self::F<A> {
        Box::pin(sa)
    }
}

impl<'a, E: 'a + Send> Fail<'a, E> for TryFutureInstance<E> {
    fn fail<A: 'a + Send>(e: E) -> Self::F<A> {
        Box::pin(async { Err(e) })
    }
}

pub struct FutureFailAny;

impl<'a> MonadFailAny<'a> for FutureFailAny {
    type W<E: 'a + Send> = TryFutureInstance<E>;

    type T = instances::future::FutureInstance;

    fn unstuff<A: 'a + Send, E: 'a + Send>(
        wa: WrapE<'a, A, E, Self>,
    ) -> Wrap<'a, Result<A, E>, Self::T> {
        wa
    }

    fn stuff<A: 'a + Send, E: 'a + Send>(
        fa: Wrap<'a, Result<A, E>, Self::T>,
    ) -> WrapE<'a, A, E, Self> {
        fa
    }

    fn map_err<A: 'a + Send, E0: 'a + Send, E1: 'a + Send>(
        wa: WrapE<'a, A, E0, Self>,
        f: impl 'a + Send + FnOnce(E0) -> E1,
    ) -> WrapE<'a, A, E1, Self> {
        Box::pin(async { wa.await.map_err(f) })
    }

    fn bind_err<A: 'a + Send, E0: 'a + Send, E1: 'a + Send>(
        wa: WrapE<'a, A, E0, Self>,
        f: impl 'a + Send + FnOnce(E0) -> WrapE<'a, A, E1, Self>,
    ) -> WrapE<'a, A, E1, Self> {
        Box::pin(async {
            match wa.await {
                Ok(a) => Ok(a),
                Err(e) => f(e).await,
            }
        })
    }

    fn bind<A: 'a + Send, B: 'a + Send, E0: 'a + Send, E1: 'a + Send>(
        wa: WrapE<'a, A, E0, Self>,
        f: impl 'a + Send + FnOnce(Result<A, E0>) -> WrapE<'a, B, E1, Self>,
    ) -> WrapE<'a, B, E1, Self> {
        Box::pin(async { f(wa.await).await })
    }

    fn rotate_out<A: 'a + Send, E0: 'a + Send, E1: 'a + Send>(
        wa: WrapE<'a, Result<A, E1>, E0, Self>,
    ) -> WrapE<'a, A, Result<E1, E0>, Self> {
        Box::pin(async {
            match wa.await {
                Ok(Ok(a)) => Ok(a),
                Ok(Err(e)) => Err(Ok(e)),
                Err(e) => Err(Err(e)),
            }
        })
    }
}
