use super::*;

/// [`Deserializer`] with a cleaner interface for parsing multiple objects
/// from one stream ("inlining").
pub trait Stream: Sized {
    /// Try to read `n` bytes. Consumes the deserializer on failure.
    fn iread_n<A, E>(
        self,
        n: usize,
        ok: impl FnOnce(&[u8]) -> A,
        err: impl FnOnce(&[u8]) -> E,
    ) -> Result<(A, Self), E>;

    /// Read all bytes, consuming the deserializer.
    fn iread_all<A>(self, ok: impl FnOnce(&[u8]) -> A) -> A;

    /// Re-export of [`Deserializer::tell`].
    fn itell(&self) -> usize;
}

/// [`StreamExt::iread_n_const`].
pub trait StreamExt<E>: Stream {
    /// `const`-length equivalent of [`Stream::iread_n`].
    fn iread_n_const<const N: usize>(
        self,
        err: impl FnOnce(&[u8]) -> E,
    ) -> Result<([u8; N], Self), E> {
        self.iread_n(
            N,
            |slice| slice.try_into().expect("`iread_n` wrong length"),
            err,
        )
    }
}

impl<I: Stream, E> StreamExt<E> for I {}

impl<D: ?Sized + Deserializer> Stream for &mut D {
    fn iread_n<A, E>(
        self,
        n: usize,
        ok: impl FnOnce(&[u8]) -> A,
        err: impl FnOnce(&[u8]) -> E,
    ) -> Result<(A, Self), E> {
        let slice = self.read_n(n);
        if slice.len() == n {
            Ok((ok(slice), self))
        } else {
            Err(err(slice))
        }
    }

    fn iread_all<A>(self, ok: impl FnOnce(&[u8]) -> A) -> A {
        ok(self.read_all())
    }

    fn itell(&self) -> usize {
        self.tell()
    }
}

/// [`StreamResultExt::seal`].
pub trait StreamResultExt<A, E> {
    /// Drop the deserializer from parsing result.
    fn seal(self) -> Result<A, E>;
}

impl<A, E, I: Stream> StreamResultExt<A, E> for Result<(A, I), E> {
    fn seal(self) -> Result<A, E> {
        self.map(|(a, _)| a)
    }
}

pub trait FromRef<T: ?Sized>: for<'a> From<&'a T> {
    fn from_ref(value: &T) -> Self {
        Self::from(value)
    }
}

impl<T: ?Sized, U: for<'a> From<&'a T>> FromRef<T> for U {}
